// $Id: SecondOrder.cpp,v 1.43 2006-08-22 01:58:35 geuzaine Exp $
//
// Copyright (C) 1997-2006 C. Geuzaine, J.-F. Remacle
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
// USA.
// 
// Please report all bugs and problems to <gmsh@geuz.org>.

#include "GModel.h"
#include "GFace.h"
#include "GEdge.h"
#include "MElement.h"
#include "MRep.h"
#include "Message.h"
#include "OS.h"

extern GModel *GMODEL;

void getEdgeVertices(GEdge *ge, MElement *ele, std::vector<MVertex*> &ve,
		     std::map<std::pair<MVertex*,MVertex*>, MVertex* > &edgeVertices,
		     bool linear)
{
  for(int i = 0; i < ele->getNumEdges(); i++){
    MEdge edge = ele->getEdge(i);
    std::pair<MVertex*, MVertex*> p(edge.getMinVertex(), edge.getMaxVertex());
    if(edgeVertices.count(p)){
      ve.push_back(edgeVertices[p]);
    }
    else{
      MVertex *v, *v0 = edge.getVertex(0), *v1 = edge.getVertex(1);
      if(linear || ge->geomType() == GEntity::DiscreteCurve){
	SPoint3 pc = edge.barycenter();
	v = new MVertex(pc.x(), pc.y(), pc.z(), ge);
      }
      else{
	double u0 = 1e6, u1 = 1e6;
	Range<double> bounds = ge->parBounds(0);
	if(ge->getBeginVertex() && ge->getBeginVertex()->mesh_vertices[0] == v0) 
	  u0 = bounds.low();
	else if(ge->getEndVertex() && ge->getEndVertex()->mesh_vertices[0] == v0) 
	  u0 = bounds.high();
	else 
	  v0->getParameter(0, u0);
	if(ge->getBeginVertex() && ge->getBeginVertex()->mesh_vertices[0] == v1) 
	  u1 = bounds.low();
	else if(ge->getEndVertex() && ge->getEndVertex()->mesh_vertices[0] == v1) 
	  u1 = bounds.high();
	else 
	  v1->getParameter(0, u1);
	if(u0 < 1e6 && u1 < 1e6){
	  double uc = 0.5 * (u0 + u1);
	  GPoint pc = ge->point(uc);
	  v = new MEdgeVertex(pc.x(), pc.y(), pc.z(), ge, uc);
	}
	else{
	  // we should normally never end up here
	  SPoint3 pc = edge.barycenter();
	  v = new MVertex(pc.x(), pc.y(), pc.z(), ge);
	}
      }
      edgeVertices[p] = v;
      ge->mesh_vertices.push_back(v);
      ve.push_back(v);
    }
  }
}

void getEdgeVertices(GFace *gf, MElement *ele, std::vector<MVertex*> &ve,
		     std::map<std::pair<MVertex*,MVertex*>, MVertex* > &edgeVertices,
		     bool linear)
{
  for(int i = 0; i < ele->getNumEdges(); i++){
    MEdge edge = ele->getEdge(i);
    std::pair<MVertex*, MVertex*> p(edge.getMinVertex(), edge.getMaxVertex());
    if(edgeVertices.count(p)){
      ve.push_back(edgeVertices[p]);
    }
    else{
      MVertex *v, *v0 = edge.getVertex(0), *v1 = edge.getVertex(1);
      if(linear || gf->geomType() == GEntity::DiscreteSurface){
	SPoint3 pc = edge.barycenter();
	v = new MVertex(pc.x(), pc.y(), pc.z(), gf);
      }
      else{
	SPoint2 p0 = gf->parFromPoint(SPoint3(v0->x(), v0->y(), v0->z()));
	SPoint2 p1 = gf->parFromPoint(SPoint3(v1->x(), v1->y(), v1->z()));
	double uc = 0.5 * (p0[0] + p1[0]);
	double vc = 0.5 * (p0[1] + p1[1]);
	GPoint pc = gf->point(uc, vc);
	v = new MFaceVertex(pc.x(), pc.y(), pc.z(), gf, uc, vc);
      }
      edgeVertices[p] = v;
      gf->mesh_vertices.push_back(v);
      ve.push_back(v);
    }
  }
}

void getEdgeVertices(GRegion *gr, MElement *ele, std::vector<MVertex*> &ve,
		     std::map<std::pair<MVertex*,MVertex*>, MVertex* > &edgeVertices,
		     bool linear)
{
  for(int i = 0; i < ele->getNumEdges(); i++){
    MEdge edge = ele->getEdge(i);
    std::pair<MVertex*, MVertex*> p(edge.getMinVertex(), edge.getMaxVertex());
    if(edgeVertices.count(p)){
      ve.push_back(edgeVertices[p]);
    }
    else{
      SPoint3 pc = edge.barycenter();
      MVertex *v = new MVertex(pc.x(), pc.y(), pc.z(), gr);
      edgeVertices[p] = v;
      gr->mesh_vertices.push_back(v);
      ve.push_back(v);
    }
  }
}

void getFaceVertices(GFace *gf, MElement *ele, std::vector<MVertex*> &vf,
		     std::map<std::vector<MVertex*>, MVertex* > &faceVertices,
		     bool linear)
{
  for(int i = 0; i < ele->getNumFaces(); i++){
    MFace face = ele->getFace(i);
    if(face.getNumVertices() != 4) continue;
    std::vector<MVertex*> p;
    face.getOrderedVertices(p);
    if(faceVertices.count(p)){
      vf.push_back(faceVertices[p]);
    }
    else{
      MVertex *v;
      if(linear || gf->geomType() == GEntity::DiscreteSurface){
	SPoint3 pc = face.barycenter();
	v = new MVertex(pc.x(), pc.y(), pc.z(), gf);
      }
      else{
	SPoint2 p0 = gf->parFromPoint(SPoint3(p[0]->x(), p[0]->y(), p[0]->z()));
	SPoint2 p1 = gf->parFromPoint(SPoint3(p[1]->x(), p[1]->y(), p[1]->z()));
	SPoint2 p2 = gf->parFromPoint(SPoint3(p[2]->x(), p[2]->y(), p[2]->z()));
	SPoint2 p3 = gf->parFromPoint(SPoint3(p[3]->x(), p[3]->y(), p[3]->z()));
	double uc = 0.25 * (p0[0] + p1[0] + p2[0] + p3[0]);
	double vc = 0.25 * (p0[1] + p1[1] + p2[1] + p3[1]);
	GPoint pc = gf->point(uc, vc);
	v = new MFaceVertex(pc.x(), pc.y(), pc.z(), gf, uc, vc);
      }
      faceVertices[p] = v;
      gf->mesh_vertices.push_back(v);
      vf.push_back(v);
    }
  }
}

void getFaceVertices(GRegion *gr, MElement *ele, std::vector<MVertex*> &vf,
		     std::map<std::vector<MVertex*>, MVertex* > &faceVertices,
		     bool linear)
{
  for(int i = 0; i < ele->getNumFaces(); i++){
    MFace face = ele->getFace(i);
    if(face.getNumVertices() != 4) continue;
    std::vector<MVertex*> p;
    face.getOrderedVertices(p);
    if(faceVertices.count(p)){
      vf.push_back(faceVertices[p]);
    }
    else{
      SPoint3 pc = face.barycenter();
      MVertex *v = new MVertex(pc.x(), pc.y(), pc.z(), gr);
      faceVertices[p] = v;
      gr->mesh_vertices.push_back(v);
      vf.push_back(v);
    }
  }
}

void setSecondOrder(GEdge *ge,
		    std::map<std::pair<MVertex*,MVertex*>, MVertex* > &edgeVertices,
		    bool linear)
{
  std::vector<MLine*> lines2;
  for(unsigned int i = 0; i < ge->lines.size(); i++){
    MLine *l = ge->lines[i];
    std::vector<MVertex*> ve;
    getEdgeVertices(ge, l, ve, edgeVertices, linear);
    lines2.push_back(new MLine2(l->getVertex(0), l->getVertex(1), ve[0]));
    delete l;
  }
  ge->lines = lines2;

  if(ge->meshRep) ge->meshRep->destroy();
}

void setSecondOrder(GFace *gf,
		    std::map<std::pair<MVertex*,MVertex*>, MVertex* > &edgeVertices,
		    std::map<std::vector<MVertex*>, MVertex* > &faceVertices,
		    bool linear)
{
  std::vector<MTriangle*> triangles2;
  for(unsigned int i = 0; i < gf->triangles.size(); i++){
    MTriangle *t = gf->triangles[i];
    std::vector<MVertex*> ve;
    getEdgeVertices(gf, t, ve, edgeVertices, linear);
    triangles2.push_back
      (new MTriangle2(t->getVertex(0), t->getVertex(1), t->getVertex(2),
		      ve[0], ve[1], ve[2]));
    delete t;
  }
  gf->triangles = triangles2;

  std::vector<MQuadrangle*> quadrangles2;
  for(unsigned int i = 0; i < gf->quadrangles.size(); i++){
    MQuadrangle *q = gf->quadrangles[i];
    std::vector<MVertex*> ve, vf;
    getEdgeVertices(gf, q, ve, edgeVertices, linear);
    getFaceVertices(gf, q, vf, faceVertices, linear);
    quadrangles2.push_back
      (new MQuadrangle2(q->getVertex(0), q->getVertex(1), q->getVertex(2),
			q->getVertex(3), ve[0], ve[1], ve[2], ve[3], vf[0]));
    delete q;
  }
  gf->quadrangles = quadrangles2;
  
  if(gf->meshRep) gf->meshRep->destroy();
}

void setSecondOrder(GRegion *gr,
		    std::map<std::pair<MVertex*,MVertex*>, MVertex* > &edgeVertices,
		    std::map<std::vector<MVertex*>, MVertex* > &faceVertices,
		    bool linear)
{
  std::vector<MTetrahedron*> tetrahedra2;
  for(unsigned int i = 0; i < gr->tetrahedra.size(); i++){
    MTetrahedron *t = gr->tetrahedra[i];
    std::vector<MVertex*> ve;
    getEdgeVertices(gr, t, ve, edgeVertices, linear);
    tetrahedra2.push_back
      (new MTetrahedron2(t->getVertex(0), t->getVertex(1), t->getVertex(2), 
			 t->getVertex(3), ve[0], ve[1], ve[2], ve[3], ve[4], ve[5]));
    delete t;
  }
  gr->tetrahedra = tetrahedra2;

  std::vector<MHexahedron*> hexahedra2;
  for(unsigned int i = 0; i < gr->hexahedra.size(); i++){
    MHexahedron *h = gr->hexahedra[i];
    std::vector<MVertex*> ve, vf;
    getEdgeVertices(gr, h, ve, edgeVertices, linear);
    getFaceVertices(gr, h, vf, faceVertices, linear);
    SPoint3 pc = h->barycenter();
    MVertex *v = new MVertex(pc.x(), pc.y(), pc.z(), gr);
    gr->mesh_vertices.push_back(v);
    hexahedra2.push_back
      (new MHexahedron2(h->getVertex(0), h->getVertex(1), h->getVertex(2), 
			h->getVertex(3), h->getVertex(4), h->getVertex(5), 
			h->getVertex(6), h->getVertex(7), ve[0], ve[1], ve[2], 
			ve[3], ve[4], ve[5], ve[6], ve[7], ve[8], ve[9], ve[10], 
			ve[11], vf[0], vf[1], vf[2], vf[3], vf[4], vf[5], v));
    delete h;
  }
  gr->hexahedra = hexahedra2;

  std::vector<MPrism*> prisms2;
  for(unsigned int i = 0; i < gr->prisms.size(); i++){
    MPrism *p = gr->prisms[i];
    std::vector<MVertex*> ve, vf;
    getEdgeVertices(gr, p, ve, edgeVertices, linear);
    getFaceVertices(gr, p, vf, faceVertices, linear);
    prisms2.push_back
      (new MPrism2(p->getVertex(0), p->getVertex(1), p->getVertex(2), 
		   p->getVertex(3), p->getVertex(4), p->getVertex(5), 
		   ve[0], ve[1], ve[2], ve[3], ve[4], ve[5], ve[6], ve[7],
		   ve[8], vf[0], vf[1], vf[2]));
    delete p;
  }
  gr->prisms = prisms2;

  std::vector<MPyramid*> pyramids2;
  for(unsigned int i = 0; i < gr->pyramids.size(); i++){
    MPyramid *p = gr->pyramids[i];
    std::vector<MVertex*> ve, vf;
    getEdgeVertices(gr, p, ve, edgeVertices, linear);
    getFaceVertices(gr, p, vf, faceVertices, linear);
    pyramids2.push_back
      (new MPyramid2(p->getVertex(0), p->getVertex(1), p->getVertex(2), 
		     p->getVertex(3), p->getVertex(4), ve[0], ve[1], ve[2], 
		     ve[3], ve[4], ve[5], ve[6], ve[7], vf[0]));
    delete p;
  }
  gr->pyramids = pyramids2;

  if(gr->meshRep) gr->meshRep->destroy();
}

template<class T>
void setFirstOrder(GEntity *e, std::vector<T*> &elements)
{
  std::vector<T*> elements1;
  for(unsigned int i = 0; i < elements.size(); i++){
    T *ele = elements[i];
    int n = ele->getNumVertices() - ele->getNumEdgeVertices() - 
      ele->getNumFaceVertices() - ele->getNumVolumeVertices();
    std::vector<MVertex*> v1;
    for(int j = 0; j < n; j++)
      v1.push_back(ele->getVertex(j));
    for(int j = n; j < ele->getNumVertices(); j++)
      ele->getVertex(j)->setVisibility(-1);
    elements1.push_back(new T(v1));
    delete ele;
  }
  elements = elements1;
  
  if(e->meshRep) e->meshRep->destroy();
}

void removeSecondOrderVertices(GEntity *e)
{
  std::vector<MVertex*> v1;
  for(unsigned int i = 0; i < e->mesh_vertices.size(); i++){
    if(e->mesh_vertices[i]->getVisibility() < 0)
      delete e->mesh_vertices[i];
    else
      v1.push_back(e->mesh_vertices[i]);
  }
  e->mesh_vertices = v1;
}

void Degre1()
{
  // replace all elements with first order elements and mark all
  // unused nodes with a -1 visibility flag
  for(GModel::eiter it = GMODEL->firstEdge(); it != GMODEL->lastEdge(); ++it){
    setFirstOrder(*it, (*it)->lines);
  }
  for(GModel::fiter it = GMODEL->firstFace(); it != GMODEL->lastFace(); ++it){
    setFirstOrder(*it, (*it)->triangles);
    setFirstOrder(*it, (*it)->quadrangles);
  }
  for(GModel::riter it = GMODEL->firstRegion(); it != GMODEL->lastRegion(); ++it){
    setFirstOrder(*it, (*it)->tetrahedra);
    setFirstOrder(*it, (*it)->hexahedra);
    setFirstOrder(*it, (*it)->prisms);
    setFirstOrder(*it, (*it)->pyramids);
  }

  // remove all nodes with a -1 visibility flag
  for(GModel::eiter it = GMODEL->firstEdge(); it != GMODEL->lastEdge(); ++it)
    removeSecondOrderVertices(*it);
  for(GModel::fiter it = GMODEL->firstFace(); it != GMODEL->lastFace(); ++it)
    removeSecondOrderVertices(*it);
  for(GModel::riter it = GMODEL->firstRegion(); it != GMODEL->lastRegion(); ++it)
    removeSecondOrderVertices(*it);
}

void Degre2()
{
  Msg(STATUS1, "Meshing second order...");
  double t1 = Cpu();

  bool linear = true;
  //bool linear = false;

  std::map<std::pair<MVertex*,MVertex*>, MVertex* > edgeVertices;
  std::map<std::vector<MVertex*>, MVertex* > faceVertices;

  // replace all elements with second order elements by creating
  // unique nodes on the edges/faces of the mesh. (To generate nodes
  // on the exact geometrical edges/faces this assumes that the
  // geometrical edges/faces are discretized with 1D/2D
  // elements. I.e., if there are only 3D elements in the mesh then
  // any new nodes on the boundary will simply be added by linear
  // interpolation.)
  for(GModel::eiter it = GMODEL->firstEdge(); it != GMODEL->lastEdge(); ++it)
    setSecondOrder(*it, edgeVertices, linear);
  for(GModel::fiter it = GMODEL->firstFace(); it != GMODEL->lastFace(); ++it)
    setSecondOrder(*it, edgeVertices, faceVertices, linear);
  for(GModel::riter it = GMODEL->firstRegion(); it != GMODEL->lastRegion(); ++it)
    setSecondOrder(*it, edgeVertices, faceVertices, linear);

  double t2 = Cpu();
  Msg(INFO, "Mesh second order complete (%g s)", t2 - t1);
  Msg(STATUS1, "Mesh");
}
